"""add_scheduable_field_to_agents

Revision ID: 9cd61b1ae70d
Revises: e35332f8d23d
Create Date: 2020-07-01 15:02:13.979828

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql.expression import true
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '9cd61b1ae70d'
down_revision = 'e35332f8d23d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('agents', sa.Column(
        'schedulable', sa.Boolean(),
        server_default=true(),
        default=True,
        nullable=False,
    ))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('agents', 'schedulable')
    # ### end Alembic commands ###
